/* artistParser.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "artistParser.h"
#include "string_utils.h"

ArtistParser::
ArtistParser(const string &name, Hint hint) {
  _name = name;

  // First, separate the name into its individual words.
  // Also mark the words that were followed by commas.
  
  int lp = 0;
  int tp = 0;
  while (tp < _name.length()) {
    if (isspace(_name[tp]) || _name[tp] == ',') {
      _words.push_back(_name.substr(lp, tp - lp));
      _comma_flags.push_back(_name[tp] == ',');
      tp++;

      while (tp < _name.length() && isspace(_name[tp])) {
	tp++;
      }
      lp = tp;
    } else {
      tp++;
    }
  }
  if (tp > lp) {
    _words.push_back(_name.substr(lp));
    _comma_flags.push_back(false);
  }

  // Look for identifiable keywords within the name, like "The" and
  // "Jr".
  _and_p = find_and(0, _words.size());
  _name_suffix_p = find_name_suffix(0, _words.size());
  _group_prefix_p = find_group_prefix(0, _words.size());

  // If we have hints, respect them.
  if (hint == H_individual) {
    get_individual_name(0, _words.size());

  } else if (hint == H_group) {
    get_group_name(0, _words.size());

    // If an article appears before the word "and", the whole thing is a
    // group name.  Period.
  } else if (_group_prefix_p >= 0 &&
      (_and_p < 0 || _group_prefix_p < _and_p)) {
    _group_prefix = _words[0];
    get_group_name(1, _words.size() - 1);

    // If we have "and" in there somewhere, that divides us into an
    // individual and a group.
  } else if (_and_p > 0) {
    get_individual_name(0, _and_p);
    get_group_name(_and_p + 1, _words.size() - (_and_p+1));

    // If the last word is a name suffix, that's a giveaway for an
    // individual.
  } else if (_name_suffix_p == _words.size() - 1) {
    get_individual_name(0, _words.size());

    // If we have exactly two words, we're probably an individual.
  } else if (_words.size() == 2) {
    get_individual_name(0, _words.size());

    // Otherwise, we're probably a group.
  } else {
    get_group_name(0, _words.size());
  }
}


int ArtistParser::
find_and(int start, int num_words) {
  for (int i = start; i < start + num_words; i++) {
    if (nocasecmp(_words[i], "and") == 0 ||
	_words[i] == "&") {
      return i;
    }
  }
  return -1;
}

int ArtistParser::
find_name_suffix(int start, int num_words) {
  for (int i = start; i < start + num_words; i++) {
    if (nocasecmp(_words[i], "jr") == 0 ||
	nocasecmp(_words[i], "jr.") == 0 ||
	nocasecmp(_words[i], "sr") == 0 ||
	nocasecmp(_words[i], "sr.") == 0 ||
	nocasecmp(_words[i], "iii") == 0) {
      return i;
    }
  }
  return -1;
}

int ArtistParser::
find_group_prefix(int start, int num_words) {
  for (int i = start; i < start + num_words; i++) {
    if (nocasecmp(_words[i], "a") == 0 ||
	nocasecmp(_words[i], "an") == 0 ||
	nocasecmp(_words[i], "the") == 0) {
      return i;
    }
  }
  return -1;
}

string ArtistParser::
assemble_words(int start, int num_words) {
  assert(start >= 0 && start < _words.size());
  assert(num_words >= 0 && start + num_words <= _words.size());
  if (num_words == 0) {
    return string();
  }

  string name;

  for (int i = start; i < start + num_words; i++) {
    name += " " + _words[i];
    if (_comma_flags[i]) {
      name += ",";
    }
  }
  
  return name.substr(1);
}

void ArtistParser::
get_group_name(int start, int num_words) {
  // Check the first word for a prefix.
  assert(num_words > 0);
  
  int prefix = find_group_prefix(start, 1);
  if (prefix >= 0) {
    _group_prefix = _words[start];
    start++;
    num_words--;
  }

  _group_name = assemble_words(start, num_words);
}

void ArtistParser::
get_individual_name(int start, int num_words) {
  assert(num_words > 0);

  // Check the last word for a suffix.
  int suffix = find_name_suffix(start + num_words - 1, 1);

  if (suffix >= 0) {
    assert(suffix == start + num_words - 1);
    _name_suffix = _words[suffix];
    num_words--;
  }

  // Is it "First Last" or "Last, First"? 
  int comma = start + num_words - 2;
  while (comma >= start && !_comma_flags[comma]) {
    comma--;
  }

  if (comma >= start) {
    // Last, First.
    _comma_flags[comma] = false;
    _last_name = assemble_words(start, comma - start + 1);
    _first_name = assemble_words(comma + 1, num_words - (comma-start+1));
  } else {
    // First Last.
    if (num_words > 1) {
      _first_name = assemble_words(start, num_words - 1);
    }
    _last_name = _words[start + num_words - 1];
  }
}    

// This output function is mainly for debugging.  
void ArtistParser::
write(ostream &out) const {
  out << "ArtistParser(" << _name << ") :"
      << "\n  _last_name = " << _last_name
      << "\n  _first_name = " << _first_name
      << "\n  _name_suffix = " << _name_suffix
      << "\n  _group_prefix = " << _group_prefix
      << "\n  _group_name = " << _group_name
      << "\n";
}

