/* playlist.h

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#ifndef PLAYLIST_H
#define PLAYLIST_H

#include "dataCell.h"
#include "tableInfo.h"
#include "referenceCount.h"

#include <gtk--.h>
#include <vector>
#include <set>

// A playlist may be a list of songs, a list of CD's, or a list of
// artists.  It consists of a number of rows, each of which
// corresponds to a single song, CD, or artist, and defines the
// primary key into the appropriate table.  

// A playlist may also contain any number of auxiliary column fields,
// which may have any header and contain any kind of data.  In
// addition to its primary key, each row defines a value for each
// auxiliary column.

class QueryResult;
class Dataset;
class PalomaDialog;
class CDReader;

class Playlist : public Gtk_Signal_Base, public ReferenceCount {
public:
  // This defines a single auxiliary column.
  class Column {
  public:
    Column(const string &name, 
	   DataCell::Type type = DataCell::DC_string, 
	   TableID foreign_table = TI_no_table);
    Column(const Column &copy);
    Column &operator = (const Column &copy);

    string format_value(const DataCell &value) const;

    string _name;
    DataCell::Type _type;
    TableID _foreign_table;
  };
  typedef vector<Column> Columns;
  typedef vector<DataCell> ColumnValues;

  // This defines a single row in the playlist.  It's a song ID and an
  // associated set of column values.
  class Row {
  public:
    Row(int primary_key, int cd_id = -1);

    int _primary_key;
    // The cd_id is known only if the song came from a particular CD.
    int _cd_id;
    ColumnValues _column_values;
  };
  typedef vector<Row> Rows;


  Playlist(TableID primary_table = TI_no_table);
  virtual ~Playlist();
  
  void append_data(const QueryResult &qr);
  
  string format_value(int r, int c) const;
  Gtk_CList *make_clist();
  Gtk_CList *get_clist() const;
  virtual void update_clist(int row);

  int get_num_rows() const;
  const Row *get_row(int row) const;
  void popup_dialog(int row);

  TableID get_primary_table() const;
  void set_owns_dialogs(bool flag);
  void set_next_form_id(int next_form_id);

  // The Playlist becomes the owner of the CDReader and will delete it
  // when it destructs.
  void set_cd_reader(CDReader *cd_reader);
  CDReader *get_cd_reader() const;

  // These virtual functions are hooks that define some peculiar
  // behavior of particular kinds of playlists (for instance, the
  // playlist of tracks listed on the CDDialog).
  virtual TableDataProxy get_table_data(int row);
  virtual void fill_dialog_data(int row, Dataset &data);
  virtual void dialog_created(int row, PalomaDialog *dialog);
  virtual void dialog_accepted(int row, PalomaDialog *dialog);

protected:
  // This class is passed into song_dialog_accepted as extra data
  // piggybacking on the SongDialog's "accept" event.  It indicates
  // the dialog pointer and the particular song number the dialog was
  // associated with.
  class AcceptDialogData {
  public: 
    AcceptDialogData(PalomaDialog *dialog, int row);
    PalomaDialog *_dialog;
    int _row;
  };

  gint got_clist_clicked(GdkEventButton *event);
  void got_dialog_accept(AcceptDialogData add);
  void got_dialog_delete(PalomaDialog *dialog);

  // We need a set of all the dialogs we've currently got open, mainly
  // so that when the playlist is deleted it can delete these too.
  typedef set<PalomaDialog *> Dialogs;
  Dialogs _dialogs;


  TableID _primary_table;
  Gtk_CList *_clist;
  CDReader *_cd_reader;

  // owns_dialogs is true if any opened child dialogs should be
  // destroyed when the Playlist object destructs.
  bool _owns_dialogs;
  int _next_form_id;

public:
  Columns _columns;
  Rows _rows;
};


#endif
