/* playlistDialog.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "playlistDialog.h"
#include "database.h"
#include "dataCell.h"
#include "songDialog.h"
#include "formDef.h"
#include "srandObject.h"
#include "widgetSize.h"
#include "exportDialog.h"

#include "next.xpm"
#include "prev.xpm"

#include <algorithm>

PlaylistDialog::
PlaylistDialog(const Dataset &data, int form_id, 
	       const PlaylistProxy &playlist) :
  PalomaDialog(data),
  _form_id(form_id),
  _playlist(playlist),
  _controls(data),
  _menu("<playlist>"),
  _shuffle_button("Shuffle"),
  _repeat_check("Repeat"),
  _shuffle_check("Reshuffle at end"),
  _songs_button("Songs"),
  _remove_button("Remove"),
  _cancel_button("Cancel")
{
  _clist = _playlist.make_clist();
  _clist->columns_autosize();
  _clist->set_reorderable(true);
  _playlist.set_owns_dialogs(false);
  _playlist.set_next_form_id
    (get_related_form_id(_form_id, _playlist.get_primary_table()));

  connect_to_method(_controls.ended, this, &playlist_ended);
  layout_window();
  setup();

  if (is_playable() && _clist->rows() > 0) {
    _clist->cause_select_row(0, 0);
  }
}

PlaylistDialog::
~PlaylistDialog() {
}

PlaylistDialog *PlaylistDialog::
make_dialog(const Dataset &data, int form_id,
	    const string &sql_command) {
  Playlist *playlist = new Playlist;
  playlist->append_data(music_db->sql_command(sql_command));
  return new PlaylistDialog(data, form_id, playlist);
}

// is_playable() returns true if the contents of the Playlist are such
// that it makes sense to have a play button, or false otherwise.
bool PlaylistDialog::
is_playable() const {
  return (_playlist.get_primary_table() == TI_songs);
}

void PlaylistDialog::
menu_get_songs() {
  if (_playlist.get_primary_table() == TI_cds) {
    // We have to do a number of sequential select statements to get
    // all the songs from the various CD's, instead of one big select
    // statement, because we want to preserve the CD's in the order
    // they are currently in the list, and we furthermore want to
    // order the tracks properly within each CD.

    Playlist *songs = new Playlist(TI_songs);
    int num_rows = _playlist.get_num_rows();
    for (int r = 0; r < num_rows; r++) {
      QueryResult qr = music_db->sql_command
	("SELECT songs.song_id, cds.cd_id, cds.title, songs.name"
	 " FROM songs, cds, cd_tracks"
	 " WHERE songs.song_id=cd_tracks.song_id"
	 " AND cd_tracks.cd_id=cds.cd_id"
	 " AND cds.cd_id=" + DataCell(_playlist.get_row(r)->_primary_key) +
	 " ORDER BY cd_tracks.track_num");
      songs->append_data(qr);
    }
    new PlaylistDialog(_data, -1, songs);

  } else if (_playlist.get_primary_table() == TI_artists) {
    Playlist *songs = new Playlist(TI_songs);
    int num_rows = _playlist.get_num_rows();
    for (int r = 0; r < num_rows; r++) {
      QueryResult qr = music_db->sql_command
	("SELECT songs.song_id, songs.performer, songs.writer, songs.name"
	 " FROM songs "
	 " WHERE songs.performer=" +
	 DataCell(_playlist.get_row(r)->_primary_key) +
	 " OR songs.writer=" +
	 DataCell(_playlist.get_row(r)->_primary_key) +
	 " ORDER BY songs.name");
      songs->append_data(qr);
    }
    new PlaylistDialog(_data, -1, songs);
  }
}

void PlaylistDialog::
menu_export() {
  int start_row = 0;
  if (_clist->selbegin() != _clist->selend()) {
    start_row = *(_clist->selbegin());
  }
  new ExportDialog(_data, _playlist, start_row, false);
}

void PlaylistDialog::
playlist_ended() {
  // The playlist has ended.  Should we reshuffle?
  if (_shuffle_check.get_active()) {
    shuffle_button_clicked();
  }

  // Should we repeat?
  if (_repeat_check.get_active() && _clist->rows() > 0) {
    _clist->cause_select_row(0, 0);
    _controls.play();
  }
}

void PlaylistDialog::
remove_button_clicked() {
  if (_clist->selbegin() != _clist->selend()) {
    int row_index = *(_clist->selbegin());
    _clist->remove_row(row_index);
  }
}

void PlaylistDialog::
cancel_button_clicked() {
  delete_self();
}

void PlaylistDialog::
shuffle_button_clicked() {
  // We shuffle the list by repeatedly swapping rows.

  _clist->freeze();
  SRandObject srand;

  // Since Gtk doesn't keep track of selections properly while
  // shuffling, we have to do it.
  int current_row =
    _clist->selbegin() == _clist->selend() ? -1 : *_clist->selbegin();

  if (current_row >= 0) {
    _clist->cause_unselect_row(current_row, 0);
  }

  int num_songs = _clist->rows();
  for (int i = 0; i < num_songs; i++) {
    int a = srand(num_songs);

    // GTK complains for some reason if we try to swap a row with
    // itself.
    if (a != i) {
      _clist->swap_rows(a, i);
      if (current_row == a) {
	current_row = i;
      } else if (current_row == i) {
	current_row = a;
      }
    }
  }

  if (current_row >= 0) {
    _clist->cause_select_row(current_row, 0);
  }

  _clist->thaw();
}

void PlaylistDialog::
layout_window() {
  switch (_playlist.get_primary_table()) {
  case TI_songs:
    set_title("Playlist");
    break;

  case TI_cds:
    set_title("Query results: CD's");
    break;

  case TI_artists:
    set_title("Query results: Artists");
    break;

  default:
    set_title("Query results");
  }

  // The menu bar.
  _menubar = _menu.get_menubar_widget("");
  assert(_menubar.get_object() != NULL);

  if (!is_playable()) {
    _menu.create_item("/File/Build a playlist of songs", "", "",
		      ItemFactoryConnector<PlaylistDialog, int>(this, &menu_get_songs));
  }
  if (_playlist.get_primary_table() == TI_songs ||
      _playlist.get_primary_table() == TI_cds) {
    _menu.create_item("/File/Export song files", "", "",
		      ItemFactoryConnector<PlaylistDialog, int>(this, &menu_export));
  }

  _menu.create_item("/File/sep", "", "<Separator>", 0);
  _menu.create_item("/File/Close", "", "",
		    ItemFactoryConnector<PlaylistDialog, int>(this, &cancel_button_clicked));

  vbox_.pack_start(*_menubar.get_object(), false, false);
  _menubar.get_object()->show();

  // The box for the main set of widgets.
  vbox_.pack_start(_box1);
  _box1.show();
  _box1.set_border_width(8);

  // The list of songs.
  _sc_window.set_policy(GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  _box1.pack_start(_sc_window, true, true, 0);
  _sc_window.show();

  _sc_window.add_with_viewport(*_clist);

  _clist->column_titles_passive();
  _clist->set_reorderable(true);
  _clist->show();

  // A halfhearted attempt to make the scrolling window the right size
  // for the CList.
  request_initial_size(_sc_window,
		       min(_clist->columns() * 240, 640),
		       min(_clist->rows() * 16 + 52, 320));


  if (is_playable()) {
    // The sound controls.
    _controls.layout(_playlist);
    _box1.pack_start(_controls, false, false, 0);
    _controls.show();

    // The Shuffle button.
    _shuffle_button.set_usize(80, 36);
    action_area_.pack_start(_shuffle_button, true, false, 0);
    _shuffle_button.set_flags(GTK_CAN_DEFAULT);
    _shuffle_button.show();
    connect_to_method(_shuffle_button.clicked, this,
		      &shuffle_button_clicked);
    
    // The check boxes.
    action_area_.pack_start(_checks, true, false, 0);
    _checks.show();
    _checks.pack_start(_repeat_check);
    _repeat_check.show();
    _checks.pack_start(_shuffle_check);
    _shuffle_check.show();

  } else {
    // The Songs button.
    _songs_button.set_usize(80, 36);
    action_area_.pack_start(_songs_button, true, false, 0);
    _songs_button.set_flags(GTK_CAN_DEFAULT);
    _songs_button.show();
    connect_to_method(_songs_button.clicked, this, &menu_get_songs);
  }

  // The Remove button.
  _remove_button.set_usize(80, 36);
  action_area_.pack_start(_remove_button, true, false, 0);
  _remove_button.set_flags(GTK_CAN_DEFAULT);
  _remove_button.show();
  connect_to_method(_remove_button.clicked, this, &remove_button_clicked);

  // The Cancel button.
  _cancel_button.set_usize(80, 36);
  action_area_.pack_start(_cancel_button, true, false, 0);
  _cancel_button.set_flags(GTK_CAN_DEFAULT);
  _cancel_button.show();
  connect_to_method(_cancel_button.clicked, this, &cancel_button_clicked);
}
