/* playerDef.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "playerDef.h"
#include "expand_variables.h"
#include "extract_words.h"
#include "format.h"
#include "string_utils.h"
#include "database.h"
#include "dataCell.h"
#include "queryResult.h"
#include "get_basename.h"
#include "protect_from_shell.h"

#include <unistd.h>
#include <fstream.h>

PlayerDef::DefMap PlayerDef::_defs;
PlayerDef::ExtensionMap PlayerDef::_extensions;
const PlayerDef *PlayerDef::_cdrom_def = NULL;


PlayerDef::StatusEntry::
StatusEntry(StatusType type, const string &literal) : 
  _type(type), 
  _literal(literal) {
}

void PlayerDef::StatusEntry::
output(ostream &out) const {
  switch (_type) {
  case PlayerDef::ST_literal:
    out << "\"" << _literal << "\"";
    break;

  case PlayerDef::ST_current_frame:
    out << "current_frame";
    break;

  case PlayerDef::ST_frames_remaining:
    out << "frames_remaining";
    break;

  case PlayerDef::ST_total_frames:
    out << "total_frames";
    break;

  case PlayerDef::ST_current_time:
    out << "current_time";
    break;

  case PlayerDef::ST_time_remaining:
    out << "time_remaining";
    break;

  case PlayerDef::ST_total_time:
    out << "total_time";
    break;

  case PlayerDef::ST_current_pos:
    out << "current_pos";
    break;

  case PlayerDef::ST_current_percent:
    out << "current_percent";
    break;

  default:
    out << "(**unexpected type**)";
  }
}

PlayerDef::
PlayerDef(const string &play_command, const string &play_status,
	  const string &verify_command, const string &verify_status) {
  _verify_command_unexpanded = trim_blanks(verify_command);
  _verify_status = expand_status(verify_status);
  _play_command_unexpanded = trim_blanks(play_command);
  _play_status = expand_status(play_status);

  // Try to determine the set of features available for the player.
  _features = 0;

  // Check the feedback string.  If any of the data we get back from
  // the player command is an indication of the current position, then
  // the player can report the current position.
  Status::const_iterator si;
  for (si = _play_status.begin(); si != _play_status.end(); ++si) {
    switch ((*si)._type) {
    case ST_current_frame:
    case ST_current_time:
    case ST_current_pos:
    case ST_current_percent:
      _features |= F_report;
      break;
    }
  }

  // Now check the start command.  For this one, we need to see what
  // variable names are referenced.  As below, we insert an
  // unlikely-to-occur string in for each variable name, and then
  // check for that string.

  VariableDef vars;

  // These are the variables that imply we can start anywhere we choose.
  vars["start_frame"] = " <:<rstart>:> ";
  vars["start_time"] = " <:<rstart>:> ";
  vars["start_pos"] = " <:<rstart>:> ";
  vars["start_percent"] = " <:<rstart>:> ";

  // These variables imply we can stop anywhere we choose.
  vars["end_frame"] = " <:<rstop>:> ";
  vars["num_frames"] = " <:<rstop>:> ";
  vars["end_time"] = " <:<rstop>:> ";
  vars["net_time"] = " <:<rstop>:> ";
  vars["end_pos"] = " <:<rstop>:> ";
  vars["end_percent"] = " <:<rstop>:> ";

  string check = expand_variables(_play_command_unexpanded, vars);

  if (check.find("<:<rstart>:>")) {
    _features |= F_rstart;
  }
  if (check.find("<:<rstop>:>")) {
    _features |= F_rstop;
  }
}

PlayerDef::
PlayerDef(const PlayerDef &copy) :
  _features(copy._features),
  _verify_command_unexpanded(copy._verify_command_unexpanded),
  _verify_status(copy._verify_status),
  _play_command_unexpanded(copy._play_command_unexpanded),
  _play_status(copy._play_status) {
}

PlayerDef &PlayerDef::
operator = (const PlayerDef &copy) {
  _features = copy._features;
  _verify_command_unexpanded = copy._verify_command_unexpanded;
  _verify_status = copy._verify_status;
  _play_command_unexpanded = copy._play_command_unexpanded;
  _play_status = copy._play_status;
  return *this;
}

PlayerDef::Status PlayerDef::
expand_status(const string &status) {
  // We need to expand a single string with variable references like
  // $current_frame into a Status object, a sequential list of things
  // to match against.  Whitespace in the input string is taken to be
  // a separator; variable names are values to look for; everything
  // else is literal text.

  // First we'll cleverly use expand_variables to ensure that each
  // variable reference is in its own word.  The replacement string is
  // chosen to be a string that is unlikely to occur in a normal
  // program output line.

  VariableDef vars;
  vars["current_frame"] = " <:<current_frame>:> ";
  vars["frames_remaining"] = " <:<frames_remaining>:> ";
  vars["total_frames"] = " <:<total_frames>:> ";
  vars["current_time"] = " <:<current_time>:> ";
  vars["time_remaining"] = " <:<time_remaining>:> ";
  vars["total_time"] = " <:<total_time>:> ";
  vars["current_pos"] = " <:<current_pos>:> ";
  vars["current_percent"] = " <:<current_percent>:> ";

  vector<string> words = extract_words(expand_variables(status, vars));

  // Now we'll go through each word and add it to the definition.
  // Each word is either literal text, or one of the above variable
  // names.

  Status result;
  vector<string>::const_iterator wi;
  for (wi = words.begin(); wi != words.end(); ++wi) {
    const string &word = (*wi);
    if (word == "<:<current_frame>:>") {
      result.push_back(StatusEntry(ST_current_frame));
    } else if (word == "<:<frames_remaining>:>") {
      result.push_back(StatusEntry(ST_frames_remaining));
    } else if (word == "<:<total_frames>:>") {
      result.push_back(StatusEntry(ST_total_frames));
    } else if (word == "<:<current_time>:>") {
      result.push_back(StatusEntry(ST_current_time));
    } else if (word == "<:<time_remaining>:>") {
      result.push_back(StatusEntry(ST_time_remaining));
    } else if (word == "<:<total_time>:>") {
      result.push_back(StatusEntry(ST_total_time));
    } else if (word == "<:<current_pos>:>") {
      result.push_back(StatusEntry(ST_current_pos));
    } else if (word == "<:<current_percent>:>") {
      result.push_back(StatusEntry(ST_current_percent));
    } else {
      result.push_back(StatusEntry(ST_literal, word));
    }
  }

  return result;
}

const PlayerDef *PlayerDef::
find_def(int type) {
  DefMap::const_iterator di;
  di = _defs.find(type);
  if (di == _defs.end()) {
    return NULL;
  }
  return &(*di).second;
}

const PlayerDef *PlayerDef::
find_def(const string &filename) {
  // Look up the filename extension in the soundfile_exts table to get
  // the file type.

  // A file may have multiple extensions.  For instance, file.wav.gz
  // has the extension .gz as well as .wav.gz.  We'll look up all of
  // them, starting with the longest string.  For this reason, we
  // can't use get_filename_extension().

  // However, we can use get_basename().
  string basename = get_basename(filename);

  int dot = basename.find('.');
  while (dot != string::npos) {
    string extension = basename.substr(dot);
    ExtensionMap::const_iterator ei;
    ei = _extensions.find(extension);
    if (ei != _extensions.end()) {
      return find_def((*ei).second);
    }

    dot = basename.find('.', dot+1);
  }

  // Nope, no match.  Too bad.
  return NULL;
}

const PlayerDef *PlayerDef::
cdrom_def() {
  return _cdrom_def;
}

void PlayerDef::
read_config() {
  // First, remove the old definitions, if any.
  _defs.clear();
  _extensions.clear();
  _cdrom_def = NULL;

  // Now get the new ones.
  QueryResult qr = music_db->sql_command
    ("SELECT soundfile_id, "
     " play_command, play_response, verify_command, verify_response"
     " FROM soundfile");

  int num_rows = qr.get_num_rows();
  int i;
  for (i = 0; i < num_rows; i++) {
    int soundfile_id = qr.get_cell(i, 0).get_int();
    string play_command = qr.get_cell(i, 1).get_string();
    string play_response = qr.get_cell(i, 2).get_string();
    string verify_command = qr.get_cell(i, 3).get_string();
    string verify_response = qr.get_cell(i, 4).get_string();
  
    _defs.insert(DefMap::value_type
		 (soundfile_id,
		  PlayerDef(play_command, play_response,
			    verify_command, verify_response)));
  }

  qr = music_db->sql_command
    ("SELECT extension, soundfile_id FROM soundfile_exts");
  num_rows = qr.get_num_rows();
  for (i = 0; i < num_rows; i++) {
    string extension = trim_blanks(qr.get_cell(i, 0).get_string());
    int soundfile_id = qr.get_cell(i, 1).get_int();

    _extensions.insert(ExtensionMap::value_type
		       (extension, soundfile_id));
  }

  qr = music_db->sql_command
    ("SELECT cdrom_type FROM configure");
  num_rows = qr.get_num_rows();
  if (num_rows > 0) {
    int cdrom_type = qr.get_cell(0, 0).get_int(-1);
    if (cdrom_type >= 0) {
      _cdrom_def = find_def(cdrom_type);
    }
  }

  /*
  DefMap::const_iterator di;
  cerr << "Definitions:\n";
  for (di = _defs.begin(); di != _defs.end(); ++di) {
    cerr << (*di).first << "\n"
	 << "  " << (*di).second._play_command_unexpanded << "\n"
	 << "  " << (*di).second._play_status << "\n"
	 << "  " << (*di).second._verify_command_unexpanded << "\n"
	 << "  " << (*di).second._verify_status << "\n";
  }
  */
}

bool PlayerDef::
has_feature(int feature_mask) const {
  return (_features & feature_mask) != 0;
}

string PlayerDef::
get_verify_command(const string &filename, int track_num) const {
  VariableDef vars;
  vars["filename"] = protect_from_shell(filename);
  vars["track"] = format_int(track_num);

  return expand_variables(_verify_command_unexpanded, vars);
}

const PlayerDef::Status &PlayerDef::
get_verify_status() const {
  return _verify_status;
}

string PlayerDef::
get_play_command(const string &sound_device,
		 const string &filename, int track_num,
		 double start_time, double end_time, 
		 int total_frames, double total_time) const {
  double start_pos, end_pos;

  if (total_time != 0.0) {
    start_pos = start_time / total_time;
    end_pos = end_time / total_time;
  } else {
    start_pos = 0.0;
    end_pos = 1.0;
  }

  int start_frame = (int)(start_pos * total_frames);
  int end_frame = (int)(end_pos * total_frames);

  VariableDef vars;
  vars["sound_device"] = sound_device;
  vars["cdrom_device"] = protect_from_shell(filename);
  vars["filename"] = protect_from_shell(filename);
  vars["track"] = format_int(track_num);
  vars["start_frame"] = format_int(start_frame);
  vars["end_frame"] = format_int(end_frame);
  vars["num_frames"] = format_int(end_frame - start_frame);
  vars["total_frames"] = format_int(total_frames);
  vars["start_time"] = format_float(start_time);
  vars["end_time"] = format_float(end_time);
  vars["net_time"] = format_float(end_time - start_time);
  vars["total_time"] = format_float(total_time);
  vars["start_pos"] = format_float(start_pos);
  vars["end_pos"] = format_float(end_pos);
  vars["start_percent"] = format_float(100.0 * start_pos);
  vars["end_percent"] = format_float(100.0 * end_pos);

  return expand_variables(_play_command_unexpanded, vars);
}

const PlayerDef::Status &PlayerDef::
get_play_status() const {
  return _play_status;
}



